/* libaudit.c -- 
 * Copyright 2004, 2005 Red Hat Inc., Durham, North Carolina.
 * All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Authors:
 *      Steve Grubb <sgrubb@redhat.com>
 *      Rickard E. (Rik) Faith <faith@redhat.com>
 */

#include "config.h"
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <time.h>
#include <pwd.h>
#include <grp.h>
#include <errno.h>
#define __USE_GNU
#include <fcntl.h>	/* O_NOFOLLOW needs gnu defined */
#undef __USE_GNU

#include "libaudit.h"


static int name_to_uid(const char *name, uid_t *uid);
static int name_to_gid(const char *name, gid_t *gid);


int audit_request_status(int fd)
{
	int rc = audit_send(fd, AUDIT_GET, NULL, 0);
	if (rc < 0)
		msg(LOG_NOTICE, "Error sending status request (%s)", 
			strerror(-rc));
	return rc;
}

int audit_set_enabled(int fd, int enabled)
{
	int rc;
	struct audit_status s;

	memset(&s, 0, sizeof(s));
	s.mask    = AUDIT_STATUS_ENABLED;
	s.enabled = enabled;
	rc = audit_send(fd, AUDIT_SET, &s, sizeof(s));
	if (rc < 0)
		msg(LOG_WARNING, "Error sending enable request (%s)", 
			strerror(-rc));
	return rc;
}

/* 
 * This function will return 0 if auditing is NOT enabled and
 * 1 if enabled, and -1 on error.
 */
int audit_is_enabled(int fd)
{
	int rc;

	if (fd < 0)
		return 0;

	if ((rc = audit_request_status(fd)) > 0) {
		struct audit_reply rep;
		int i;
		int timeout = 30; /* tenths of seconds */

	        for (i = 0; i < timeout; i++) {
			struct timespec t;
			t.tv_sec  = 0;
			t.tv_nsec = 100000000;
			nanosleep(&t, NULL);

			rc = audit_get_reply(fd, &rep, GET_REPLY_NONBLOCKING,0);
			if (rc > 0) {
                	        /* If we get done or error, break out */
                        	if (rep.type == NLMSG_DONE || 
					rep.type == NLMSG_ERROR)
	                                break;

        	                /* If its not status, keep looping */
	                        if (rep.type != AUDIT_GET)
        	                        continue;

				/* Found it... */
				return rep.status->enabled;
			}
		}
	}
	if (rc == -EPERM) {
		/* If we get this, then the kernel supports auditing
		 * but we don't have enough privelege to write to the
		 * socket. Therefore, we have already been authenticated
		 * and we are a common user. Just act as though auditing
		 * is not enabled. Any other error we take seriously.
		 */
		return 0;
	}
	return -1;
}

int audit_set_failure(int fd, int failure)
{
	int rc;
	struct audit_status s;

	memset(&s, 0, sizeof(s));
	s.mask    = AUDIT_STATUS_FAILURE;
	s.failure = failure;
	rc = audit_send(fd, AUDIT_SET, &s, sizeof(s));
	if (rc < 0)
		msg(LOG_WARNING, "Error sending failure mode request (%s)", 
			strerror(-rc));
	return rc;
}

/*
 * This function returns -1 on error and 1 on success.
 */
int audit_set_pid(int fd, int pid)
{
	struct audit_status s;
	struct audit_reply rep;
	struct timespec t;
	int rc;

	memset(&s, 0, sizeof(s));
	s.mask    = AUDIT_STATUS_PID;
	s.pid     = pid;
	rc = audit_send(fd, AUDIT_SET, &s, sizeof(s));
	if (rc < 0) {
		msg(LOG_WARNING, "Error sending failure mode request (%s)", 
			strerror(-rc));
		return rc;
	}

	/* Now we'll see if there's any reply message. This only
           happens on error. It is not fatal if there is no message.
	   As a matter of fact, we don't do anything with the message
	   besides gobble it. */
	t.tv_sec  = 0;
	t.tv_nsec = 100000000;	/* .1 second */
	nanosleep(&t, NULL);
	(void)audit_get_reply(fd, &rep, GET_REPLY_NONBLOCKING, 0);
	return 1;
}

int audit_set_rate_limit(int fd, int limit)
{
	int rc;
	struct audit_status s;

	memset(&s, 0, sizeof(s));
	s.mask       = AUDIT_STATUS_RATE_LIMIT;
	s.rate_limit = limit;
	rc = audit_send(fd, AUDIT_SET, &s, sizeof(s));
	if (rc < 0)
		msg(LOG_WARNING, "Error sending rate limit request (%s)", 
			strerror(-rc));
	return rc;
}

int audit_set_backlog_limit(int fd, int limit)
{
	int rc;
	struct audit_status s;

	memset(&s, 0, sizeof(s));
	s.mask          = AUDIT_STATUS_BACKLOG_LIMIT;
	s.backlog_limit = limit;
	rc = audit_send(fd, AUDIT_SET, &s, sizeof(s));
	if (rc < 0)
		msg(LOG_WARNING, "Error sending backlog limit request (%s)", 
			strerror(-rc));
	return rc;
}

int audit_request_list(int fd)
{
	int rc = audit_send(fd, AUDIT_LIST, NULL, 0);
	if (rc < 0)
		msg(LOG_WARNING, "Error sending list request (%s)", 
			strerror(-rc));
	return rc;
}

int audit_insert_watch(int fd, struct audit_watch *req)
{
	int rc = audit_send(fd, AUDIT_WATCH_INS, req, sizeof(*req));
	if (rc < 0)
		msg(LOG_WARNING, "Error sending watch insert request (%s)", 
			strerror(-rc));
	return rc;
}

int audit_remove_watch(int fd, struct audit_watch *req)
{
	int rc = audit_send(fd, AUDIT_WATCH_REM, req, sizeof(*req));
	if (rc < 0)
		msg(LOG_WARNING, "Error sending watch remove request (%s)", 
			strerror(-rc));
	return rc;
}

int audit_add_rule(int fd, struct audit_rule *rule, int flags, int action)
{
	int rc;

	rule->flags  = flags;
	rule->action = action;
	rc = audit_send(fd, AUDIT_ADD, rule, sizeof(*rule));
	if (rc < 0)
		msg(LOG_WARNING, "Error sending add rule request (%s)", 
			strerror(-rc));
	return rc;
}

int audit_delete_rule(int fd, struct audit_rule *rule, int flags, int action)
{
	int rc;

	rule->flags  = flags;
	rule->action = action;
	rc = audit_send(fd, AUDIT_DEL, rule, sizeof(*rule));
	if (rc < 0)
		msg(LOG_WARNING, "Error sending delete rule request (%s)", 
			strerror(-rc));
	return rc;
}


/*
 * This function will retreive the loginuid or -1 if there
 * is an error.
 */
uid_t audit_getloginuid(void)
{
	uid_t uid;
	char fn[PATH_MAX+1];
	int len, in;

	snprintf(fn, sizeof(fn), "/proc/%d/loginuid", getpid());
	in = open(fn, O_NOFOLLOW|O_RDONLY|O_NONBLOCK);
	if (in < 0)
		return -1;
	len = read(in, fn, PATH_MAX);
	if (len < 0 || len > PATH_MAX)
		return -1;
	fn[len] = 0;
	errno = 0;
	uid = strtol(fn, 0, 10);
	if (errno)
		return -1;
	else
		return uid;
}

int audit_send_message(int fd, const char *message)
{
	if (fd >= 0) {
		int rc = audit_send(fd, AUDIT_USER, message, strlen(message)+1);
		if (rc < 0)
			msg(LOG_WARNING, 
				"Error sending user message request (%s)", 
				strerror(-rc));
		return rc;
	}
	else { /* If for some reason fd < 0, send to syslog */
// FIXME: This should be configurable. Either syslog it or exit.
		char buff[1024];
		snprintf(buff, sizeof(buff), 
			"pid=%d uid=%d loginuid=%d message=%s", 
			getpid(), getuid(), audit_getloginuid(), message);
		syslog(LOG_ERR, "Can't send to audit system: AUDIT_USER %s", 
				buff);
		return 0;
	}
}

/*
 * This function will send a message to the kernel only if
 * the audit system is enabled. It returns the sequence number
 * which is > 0 on success and 0 or less on error.
 */
int audit_send_message_if_enabled(int fd, const char *message)
{
	if (audit_is_enabled(fd) > 0)
		return audit_send(fd, AUDIT_USER, message, strlen(message)+1);
	else
		return 0;
}

/*
 * This function is used to send avc denial messages from user
 * space back into the kernel.
 */
int audit_vlog(int fd, const char *fmt, va_list ap)
{
	char buffer[1024];

	vsnprintf(buffer, sizeof(buffer), fmt, ap);
	return audit_send_message(fd, buffer);
}

/*
 * This function will log a message to the kernel only if
 * the audit system is enabled. It returns the sequence number
 * which is > 0 on success and 0 or less on error.
 */
int audit_vlog_if_enabled(int fd, const char *fmt, va_list ap)
{
	if (audit_is_enabled(fd) > 0) {
		char buffer[1024];

		vsnprintf(buffer, sizeof(buffer), fmt, ap);
		return audit_send_message(fd, buffer);
	}
	return 0;
}

int audit_log(int fd, const char *fmt, ...)
{
	int rc;
	va_list ap;

	va_start(ap, fmt);
	rc = audit_vlog(fd, fmt, ap);
	va_end(ap);

	return rc;
}

/*
 * This function will log a message to the kernel only if
 * the audit system is enabled. It returns the sequence number
 * which is > 0 on success and 0 or less on error.
 */
int audit_log_if_enabled(int fd, const char *fmt, ...)
{
	if (audit_is_enabled(fd) > 0) {
		int rc;
		va_list ap;

		va_start(ap, fmt);
		rc = audit_vlog(fd, fmt, ap);
		va_end(ap);

		return rc;
	}
	else
		return 0;
}

/*
 * This function returns 0 on success and 1 on failure
 */
int audit_set_loginuid(uid_t uid)
{
	char fn[PATH_MAX];
	char loginuid[16];
	int o, count, rc = 0;

	memset(loginuid, 0, sizeof(loginuid));
	count = snprintf(loginuid, sizeof(loginuid), "%d", uid);
	snprintf(fn, sizeof(fn), "/proc/%d/loginuid", getpid());
	o = open(fn, O_NOFOLLOW|O_WRONLY|O_TRUNC);
	if (o >= 0) {
	        int block, offset = 0;
                                                                                
        	while (count > 0) {
			block = write(o, &loginuid[offset], count);

			if (block < 0) {
				if (errno == EINTR)
					continue;

				msg(LOG_ERR, "Error writing loginuid to %s", 
					fn);
				return 1;
			}
			offset += block;
			count -= block;
		}
		close(o); 
	}
	else {
		msg(LOG_ERR, "Error opening %s to record loginuid", fn);
		rc = 1;
	}
	return rc;
}

struct audit_rule *audit_rule_alloc(void)
{
	struct audit_rule *rule;

	if (!(rule = malloc(sizeof(*rule))))
		return NULL;
	memset(rule, 0, sizeof(*rule));
	return rule;
}

int audit_rule_syscall(struct audit_rule *rule, int syscall)
{
	int word = AUDIT_WORD(syscall);
	int bit  = AUDIT_BIT(syscall);

	if (word > AUDIT_BITMASK_SIZE) 
		return -1;
	rule->mask[word] |= bit;
	return 0;
}

int audit_rule_syscallbyname(struct audit_rule *rule,
                             const char *syscall)
{
	int nr;
	int i;

	if (!strcmp(syscall, "all")) {
		for (i = 0; i < AUDIT_BITMASK_SIZE; i++) 
			rule->mask[i] = ~0;
		return 0;
	}
	nr = audit_name_to_syscall(syscall);
	if (nr < 0) {
		if (isdigit(syscall[0]))
			nr = strtol(syscall, NULL, 0);
	}
	if (nr >= 0) 
		return audit_rule_syscall(rule, nr);
	return -1;
}

int audit_rule_field(struct audit_rule *rule, int field,
                     unsigned int value)
{
	if (rule->field_count >= AUDIT_MAX_FIELDS) 
		return -1;
	rule->fields[rule->field_count] = field;
	rule->values[rule->field_count] = value;
	++rule->field_count;
	return 0;
}

int audit_rule_fieldpair(struct audit_rule *rule, const char *pair)
{
	char       buf[128];
	const char *f = pair;
	const char *v = strchr(pair, '=');
	int        field;
	int        negated = 0;
    
	if (!f || !v || f == v)
		return -1;

	if (v > pair && v[-1] == '!')
        	negated = 1;

	snprintf(buf, sizeof(buf), "%*.*s", v-f-negated, v-f-negated, f);
	msg(LOG_DEBUG,"buf=%s\n", buf);
	if ((field = audit_name_to_field(buf)) < 0) 
		return -2;
	msg(LOG_DEBUG,"f%d%s%s\n", field, negated ? "!=" : "=", v+1);
	rule->fields[rule->field_count] = field | (negated ? AUDIT_NEGATE : 0);
	switch (field)
	{
		case AUDIT_UID:
		case AUDIT_EUID:
		case AUDIT_SUID:
		case AUDIT_FSUID:
		case AUDIT_LOGINUID:
			if (isdigit((char)*(v+1))) 
				rule->values[rule->field_count] = 
					strtol(v+1, NULL, 0);
			else {
				if (name_to_uid(v+1, 
					&rule->values[rule->field_count])) {
					msg(LOG_ERR, "Unknown user: %s", pair);
					return -2;
				}
			}
			break;
		case AUDIT_GID:
		case AUDIT_EGID:
		case AUDIT_SGID:
		case AUDIT_FSGID:
			if (isdigit((char)*(v+1))) 
				rule->values[rule->field_count] = 
					strtol(v+1, NULL, 0);
			else {
				if (name_to_gid(v+1, 
					&rule->values[rule->field_count])) {
					msg(LOG_ERR, "Unknown group: %s", pair);
					return -2;
				}
			}
			break;
		default:
			rule->values[rule->field_count] = strtol(v+1, NULL, 0);
			break;
	}
	++rule->field_count;
	return 0;
}

void audit_rule_free(struct audit_rule *rule)
{
	if (rule) 
		free(rule);
}

/*
* This function will take an audit structure and return a
* text buffer that's unformatted for writing to disk. If there 
* is an error the return value is NULL. It is the caller's 
* responsibility free the buffer.
*/
const char *format_raw(const struct audit_reply *rep)
{
	char *buf, *ptr;
	int max_buf=8192; //8K

	if (rep==NULL)
		return NULL;

	ptr=buf=malloc((max_buf)*sizeof(char));
	if (buf==NULL)
		return NULL;

	snprintf(buf, max_buf, "type=%s msg=%.*s", 
		audit_msg_type_to_name(rep->type),
		rep->len, rep->message);

	/* Replace \n with space so it looks nicer. */
	while ((ptr = strchr(ptr, 0x0A)) != NULL)
		*ptr = ' ';

	return buf;
}

static int name_to_uid(const char *name, uid_t *uid)
{
	struct passwd *pw;

	pw = getpwnam(name);
	if (pw == NULL) 
		return 1;

	memset(pw->pw_passwd, ' ', strlen(pw->pw_passwd));
	*uid = pw->pw_uid;
	return 0;
}

static int name_to_gid(const char *name, gid_t *gid)
{
	struct group *gr;

	gr = getgrnam(name);
	if (gr == NULL) 
		return 1;
 
	*gid = gr->gr_gid;
	return 0;
}

