/* libaudit.h -- 
 * Copyright 2004, 2005 Red Hat Inc., Durham, North Carolina.
 * All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Authors:
 *	Steve Grubb <sgrubb@redhat.com>
 *	Rickard E. (Rik) Faith <faith@redhat.com>
 */
#ifndef _LIBAUDIT_H_
#define _LIBAUDIT_H_

#include <asm/types.h>
#include <stdint.h>
#include <sys/socket.h>
#include <linux/netlink.h>
#include <linux/audit.h>
#include <stdarg.h>
#include <syslog.h>
#include <linux/fs.h>	/* For watch */

/* TODO: Add this as a kernel patch to audit.h */
/* These audit message types are for audit daemon local use. */
#define AUDIT_DAEMON		1500	/* Define daemon local records */
#ifndef AUDIT_WATCH_INS
#define AUDIT_WATCH_INS		1007	/* Insert file/dir watch entry */
#define AUDIT_WATCH_REM		1008	/* Remove file/dir watch entry */
struct audit_watch {
	int     namelen;
	int     fklen;
	char    *name;
	char    *filterkey;
	uint32_t perms;
};
/* 32 byte max key size */
#define AUDIT_FILTERKEY_MAX    32

#endif

/* defines for audit subsystem */

struct audit_reply {
	struct audit_message msg;
	int                  type;
	int                  len;
	struct nlmsghdr      *nlh;
	struct audit_status  *status;
	struct audit_rule    *rule;
	struct audit_login   *login;
	const char           *message;
	struct nlmsgerr      *error;
	int                   watch;
};

struct auditd_reply_list {
	struct audit_reply reply;
	struct auditd_reply_list *next;
};

/* Messages */
typedef enum { MSG_STDERR, MSG_SYSLOG } message_t;
typedef enum { DBG_NO, DBG_YES } debug_message_t;
void set_message_mode(message_t mode, debug_message_t debug);
void msg(int priority, const char *fmt, ...)
#ifdef __GNUC__
	__attribute__ ((format (printf, 2, 3)));
#else
	;
#endif

/* General */
typedef enum { GET_REPLY_BLOCKING=0, GET_REPLY_NONBLOCKING } reply_t;
extern int  audit_open(void);
extern void audit_close(int fd);
extern int  audit_get_reply(int fd, struct audit_reply *rep, reply_t block, 
		int peek);
extern int audit_send(int fd, int type, const void *data, int size);
extern uid_t audit_getloginuid(void);

/* Translation functions */
extern int        audit_name_to_field(const char *field);
extern const char *audit_field_to_name(int field);
extern int        audit_name_to_syscall(const char *syscall);
extern const char *audit_syscall_to_name(int syscall);
extern int        audit_name_to_flag(const char *flag);
extern const char *audit_flag_to_name(int flag);
extern int        audit_name_to_action(const char *action);
extern const char *audit_action_to_name(int action);
extern int audit_name_to_msg_type(const char *msg_type);
extern const char *audit_msg_type_to_name(int msg_type);

/* AUDIT_GET */
extern int  audit_request_status(int fd);
extern int audit_is_enabled(int fd);

/* AUDIT_SET */
extern int  audit_set_enabled(int fd, int enabled);
extern int  audit_set_failure(int fd, int failure);
extern int  audit_set_pid(int fd, int pid);
extern int  audit_set_rate_limit(int fd, int limit);
extern int  audit_set_backlog_limit(int fd, int limit);

/* AUDIT_LIST */
extern int  audit_request_list(int fd);

/* AUDIT_WATCH */
extern int audit_insert_watch(int fd, struct audit_watch *req);
extern int audit_remove_watch(int fd, struct audit_watch *req);

/* AUDIT_ADD */
extern int  audit_add_rule(int fd, struct audit_rule *rule,
                           int flags, int action);

/* AUDIT_DEL */
extern int  audit_delete_rule(int fd, struct audit_rule *rule,
                              int flags, int action);

/* AUDIT_USER */
extern int audit_vlog(int fd, const char *fmt, va_list ap);
extern int audit_vlog_if_enabled(int fd, const char *fmt, va_list ap);
extern int audit_log(int fd, const char *fmt, ...);
extern int audit_log_if_enabled(int fd, const char *fmt, ...);
extern int audit_send_message(int fd, const char *message);
extern int audit_send_message_if_enabled(int fd, const char *message);

/* AUDIT_LOGIN */
extern int  audit_set_loginuid(uid_t uid);

/* Rule-building helper functions */
extern struct audit_rule *audit_rule_alloc(void);
extern int  audit_rule_syscall(struct audit_rule *rule, int syscall);
extern int  audit_rule_syscallbyname(struct audit_rule *rule,
                                     const char *syscall);
extern int  audit_rule_field(struct audit_rule *rule, int field,
                             unsigned int value);
extern int  audit_rule_fieldpair(struct audit_rule *rule, const char *pair);
extern void audit_rule_free(struct audit_rule *rule);

/* Formatting functions */
extern const char *format_raw(const struct audit_reply *rep);

#endif
