/* lookup_table.c -- 
 * Copyright 2004, 2005 Red Hat Inc., Durham, North Carolina.
 * All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Authors:
 *      Steve Grubb <sgrubb@redhat.com>
 *      Rickard E. (Rik) Faith <faith@redhat.com>
 */

#include "config.h"
#include <string.h>
#include <unistd.h>

#include "libaudit.h"
#define __KERNEL__
#include <sys/syscall.h>
#undef __KERNEL__

struct transtab {
    int        value;
    const char *name;
};

static struct transtab syscalltab[] = {
    #include "syscalltab.h"
};
#define AUDIT_SYSCALL_NAMES (sizeof(syscalltab)/sizeof(syscalltab[0]))

static struct transtab fieldtab[] = {
    { AUDIT_PID,      "pid"      },
    { AUDIT_UID,      "uid"      },
    { AUDIT_EUID,     "euid"     },
    { AUDIT_SUID,     "suid"     },
    { AUDIT_FSUID,    "fsuid"    },
    { AUDIT_GID,      "gid"      },
    { AUDIT_EGID,     "egid"     },
    { AUDIT_SGID,     "sgid"     },
    { AUDIT_FSGID,    "fsgid"    },
    { AUDIT_LOGINUID, "loginuid" },
    { AUDIT_PERS,     "pers"     },

    { AUDIT_DEVMAJOR, "devmajor" },
    { AUDIT_DEVMINOR, "devminor" },
    { AUDIT_INODE,    "inode"    },
    { AUDIT_EXIT,     "exit"     },
    { AUDIT_SUCCESS,  "success"  },
    { AUDIT_ARG0,     "a0"       },
    { AUDIT_ARG1,     "a1"       },
    { AUDIT_ARG2,     "a2"       },
    { AUDIT_ARG3,     "a3"       },
};
#define AUDIT_FIELD_NAMES (sizeof(fieldtab)/sizeof(fieldtab[0]))

static struct transtab flagtab[] = {
    { AUDIT_PER_TASK, "task"     },
    { AUDIT_AT_ENTRY, "entry"    },
    { AUDIT_AT_EXIT,  "exit"     }
};
#define AUDIT_FLAG_NAMES (sizeof(flagtab)/sizeof(flagtab[0]))

static struct transtab actiontab[] = {
    { AUDIT_NEVER,    "never"    },
    { AUDIT_POSSIBLE, "possible" },
    { AUDIT_ALWAYS,   "always"   }
};
#define AUDIT_ACTION_NAMES (sizeof(actiontab)/sizeof(actiontab[0]))

/* 
 * This table is arranged based on statistics so more frequent items
 * are at the top.
 */
static struct transtab msg_typetab[] = {
    { AUDIT_KERNEL,  "KERNEL" },
    { AUDIT_USER,    "USER"   },
    { AUDIT_DAEMON,  "DAEMON" },
    { AUDIT_LOGIN,   "LOGIN"  },
    { AUDIT_SET,     "SET"    },
    { AUDIT_ADD,     "ADD"    },
    { AUDIT_LIST,    "LIST"   },
    { AUDIT_DEL,     "DEL"    },
    { AUDIT_GET,     "GET"    }
};
#define AUDIT_MSG_TYPE_NAMES (sizeof(msg_typetab)/sizeof(msg_typetab[0]))

static int audit_lookup_name(struct transtab *table, int length,
                             const char *name)
{
    int i;
    
    for (i = 0; i < length; i++)
        if (!strcmp(table[i].name, name)) return table[i].value;
    return -1;
}

static const char *audit_lookup_number(struct transtab *table, int length,
                                       int number)
{
    int i;
    
    for (i = 0; i < length; i++)
        if (table[i].value == number) return table[i].name;
    return NULL;
}

int audit_name_to_field(const char *field)
{
    return audit_lookup_name(fieldtab, AUDIT_FIELD_NAMES, field);
}

const char *audit_field_to_name(int field)
{
    return audit_lookup_number(fieldtab, AUDIT_FIELD_NAMES, field);
}

int audit_name_to_syscall(const char *syscall)
{
    return audit_lookup_name(syscalltab, AUDIT_SYSCALL_NAMES, syscall);
}

const char *audit_syscall_to_name(int syscall)
{
    return audit_lookup_number(syscalltab, AUDIT_SYSCALL_NAMES, syscall);
}

int audit_name_to_flag(const char *flag)
{
    return audit_lookup_name(flagtab, AUDIT_FLAG_NAMES, flag);
}

/* FIXME: only returns the first one in the list. */
const char *audit_flag_to_name(int flag)
{
	unsigned int i;
    
	for (i = 0; i < AUDIT_FLAG_NAMES; i++)
		if (flagtab[i].value & flag)
			return flagtab[i].name;
	return NULL;
}

int audit_name_to_action(const char *action)
{
    return audit_lookup_name(actiontab, AUDIT_ACTION_NAMES, action);
}

const char *audit_action_to_name(int action)
{
    return audit_lookup_number(actiontab, AUDIT_ACTION_NAMES, action);
}

int audit_name_to_msg_type(const char *msg_type)
{
    return audit_lookup_name(msg_typetab, AUDIT_MSG_TYPE_NAMES, msg_type);
}

const char *audit_msg_type_to_name(int msg_type)
{
    return audit_lookup_number(msg_typetab, AUDIT_MSG_TYPE_NAMES, msg_type);
}

