/* netlink.c --
 * Copyright 2004, 2005 Red Hat Inc., Durham, North Carolina.
 * All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Authors:
 *      Steve Grubb <sgrubb@redhat.com>
 *      Rickard E. (Rik) Faith <faith@redhat.com>
 */

#include "config.h"
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include "libaudit.h"

#ifndef NETLINK_AUDIT
#define NETLINK_AUDIT 9
#endif

static int adjust_reply(struct audit_reply *rep, int len);
static int check_ack(int fd, int seq);

/*
 * This function opens a connection to the kernel's audit
 * subsystem. You must be root for the call to succeed. On error,
 * a negative value is returned. On success, the file descriptor is
 * returned - which can be 0 or higher.
 */
int audit_open(void)
{
	struct sockaddr_nl addr;
	int rc, fd = socket(PF_NETLINK, SOCK_RAW, NETLINK_AUDIT);

	if (fd < 0) {
		msg(LOG_ERR, "Error creating netlink socket (%s)", 
			strerror(errno));
		return fd;
	}

	addr.nl_family = AF_NETLINK;
	addr.nl_pid = 0;
	addr.nl_groups = 0;
	rc = bind(fd, (struct sockaddr*)&addr, sizeof(addr));
	if (rc < 0) {
		close(fd);
		msg(LOG_ERR, "Error binding netlink socket (%s)", 
			strerror(errno));
		return -1;
	}
	if (fcntl(fd, F_SETFD, FD_CLOEXEC) == -1) {
		close(fd);
		msg(LOG_ERR, "Error setting netlink socket CLOEXEC flag (%s)", 
		strerror(errno));
		return -1;
	}
	return fd;
}


void audit_close(int fd)
{
	if (fd >= 0)
		close(fd);
}


/*
 * This function returns -1 on error, 0 if error response received,
 * and > 0 if packet OK.
 */
int audit_get_reply(int fd, struct audit_reply *rep, reply_t block, int peek)
{
	int len;
	struct sockaddr_nl nladdr;
	socklen_t nladdrlen = sizeof(nladdr);

	if (fd < 0)
		return -1;

	if (block == GET_REPLY_NONBLOCKING)
		block = MSG_DONTWAIT;

retry:
	len = recvfrom(fd, &rep->msg, sizeof(rep->msg), block|peek,
		(struct sockaddr*)&nladdr, &nladdrlen);

	if (len < 0) {
		if (errno == EINTR)
			goto retry;
		if (errno != EAGAIN)
			msg(LOG_ERR, "Error receiving netlink packet (%s)", 
				strerror(errno));
		return len;
	}
	if (nladdrlen != sizeof(nladdr)) {
		msg(LOG_ERR, "Bad address size reading netlink socket");
		return -1;
	}
	if (nladdr.nl_pid) {
		msg(LOG_ERR, "Spoofed packet received on netlink socket");
		return -1;
	}

	return adjust_reply(rep, len); 
}


/* 
 * This function returns 0 on error and len on success.
 */
static int adjust_reply(struct audit_reply *rep, int len)
{
	rep->type    = rep->msg.nlh.nlmsg_type;
	rep->len     = rep->msg.nlh.nlmsg_len;
	rep->nlh     = &rep->msg.nlh;
	rep->status  = NULL;
	rep->rule    = NULL;
	rep->message = NULL;
	rep->error   = NULL;
	rep->watch   = 0;
	if (!NLMSG_OK(rep->nlh, (unsigned int)len))
		return 0;
	switch (rep->type) {
		case NLMSG_ERROR: 
			rep->error   = NLMSG_DATA(rep->nlh); 
			break;
		case AUDIT_GET:   
			rep->status  = NLMSG_DATA(rep->nlh); 
			break;
		case AUDIT_LIST:  
			rep->rule    = NLMSG_DATA(rep->nlh); 
			break;
		case AUDIT_LOGIN:
		case AUDIT_KERNEL:
		case AUDIT_USER:  
			rep->message = NLMSG_DATA(rep->nlh); 
			break;
		case AUDIT_WATCH_INS:
		case AUDIT_WATCH_REM:
			memcpy(&rep->watch, NLMSG_DATA(rep->nlh), sizeof(int));
			break;
	}
	return len;
}


/*
 *  Return values:   success: positive non-zero sequence number
 *                   error:   -errno
 *                   short:   0
 */
int audit_send(int fd, int type, const void *data, int size)
{
	static int sequence = 0;
	struct audit_message req;
	int retval;
	struct sockaddr_nl addr;

	/* Due to user space library callbacks, there's a chance that
	   a -1 for the fd could be passed. Just check for and handle it. */
	if (fd < 0) {
		errno = EBADF;
		return -errno;
	}

	if (++sequence < 0) 
		sequence = 1;

	memset(&req, 0, sizeof(req));
	req.nlh.nlmsg_len = NLMSG_ALIGN(req.nlh.nlmsg_len) + NLMSG_SPACE(size);
	req.nlh.nlmsg_type = type;
	req.nlh.nlmsg_flags = NLM_F_REQUEST|NLM_F_ACK;
	req.nlh.nlmsg_seq = sequence;
	if (size)
		memcpy(NLMSG_DATA(&req.nlh), data, size);
	addr.nl_family = AF_NETLINK;
	addr.nl_pid = 0;
	addr.nl_groups = 0;

	retval = sendto(fd, &req, req.nlh.nlmsg_len, 0,
		(struct sockaddr*)&addr, sizeof(addr));
	if (retval == (int)req.nlh.nlmsg_len) {
		if ((retval = check_ack(fd, sequence)) == 0)
			return sequence;
		else
			return -retval; 
	}
	if (retval < 0) {
		msg(LOG_WARNING,"Error sending netlink packet (%s)\n", 
			strerror(errno));
		return -errno;
	}
	return 0;
}

#include <time.h>
static int check_ack(int fd, int seq)
{
	int rc;
	struct audit_reply rep;
	struct timespec t;

	t.tv_sec  = 0;
	t.tv_nsec = 100000000; /* .1 second */
	nanosleep(&t, NULL);

	rc = audit_get_reply(fd, &rep, GET_REPLY_NONBLOCKING, MSG_PEEK);
	if (rep.type == NLMSG_ERROR && rep.error->error) {
		return -rep.error->error;
	}
	return 0;
}

