/* auditd-event.c -- 
 * Copyright 2004, 2005 Red Hat Inc., Durham, North Carolina.
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Authors:
 *   Steve Grubb <sgrubb@redhat.com>
 *
 */

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include <signal.h>
#define __USE_GNU
#include <fcntl.h>	/* O_NOFOLLOW needs gnu defined */
#undef __USE_GNU
#include <sys/stat.h>
#include <errno.h>
#include <string.h>
#include <sys/vfs.h>
#include "auditd-event.h"
#include "libaudit.h"

/* This is defined in auditd.c */
extern volatile int stop;

struct auditd_consumer_data {
    struct daemon_conf *config;
    pthread_mutex_t queue_lock;
    pthread_cond_t queue_nonempty;
    struct auditd_reply_list *head;
    struct auditd_reply_list *tail;
    int log_fd;
    FILE *log_file;
};

/* Local function prototypes */
static void *event_thread_main(void *arg); 
static int write_to_log(const char *buf, struct auditd_consumer_data *data);
static void check_space_left(int lfd, struct daemon_conf *config);
void do_space_left_action(failure_action_t action);
void do_disk_full_action(struct daemon_conf *config);
static void rotate_logs(struct auditd_consumer_data *data);
static int open_audit_log(struct auditd_consumer_data *data);


/* Local Data */
static struct auditd_consumer_data consumer_data;
static pthread_t event_thread;
static int fs_space_warning = 0;
static int fs_admin_space_warning = 0;
static int fs_space_left = 1;
static int logging_suspended = 0;

void shutdown_events(void)
{
	/* Give it 5 seconds to clear the queue */
	alarm(5);
	pthread_join(event_thread, NULL);	
}

int init_event(struct daemon_conf *config)
{
	/* Store the netlink descriptor and config info away */
	consumer_data.config = config;

	/* Setup IPC mechanisms */
	pthread_mutex_init(&consumer_data.queue_lock, NULL);
	pthread_cond_init(&consumer_data.queue_nonempty, NULL);

	/* Reset the queue */
	consumer_data.head = consumer_data.tail = NULL;

	/* Now open the log */
	if (config->daemonize == D_BACKGROUND) {
		if (open_audit_log(&consumer_data))
			return 1;
	} else {
		consumer_data.log_fd = 1; // stdout
		consumer_data.log_file = fdopen(consumer_data.log_fd, "a");
		if (consumer_data.log_file == NULL) {
			msg(LOG_ERR, "Error setting up stdout descriptor (%s)", 
				strerror(errno));
			return 1;
		}
		/* Set it to line buffering */
		setlinebuf(consumer_data.log_file);
	}

	/* Create the worker thread */
	if (pthread_create(&event_thread, NULL,
			event_thread_main, &consumer_data) < 0) {
		msg(LOG_ERR, "Couldn't create consumer thread, exiting");
		fclose(consumer_data.log_file);
		return 1;
	}

	if (config->daemonize == D_BACKGROUND)
		check_space_left(consumer_data.log_fd, config);
	return 0;
}


void enqueue_event(struct auditd_reply_list *rep)
{
	pthread_mutex_lock(&consumer_data.queue_lock);

	if (consumer_data.head == NULL) {
		consumer_data.head = consumer_data.tail = rep;
		rep->next = NULL;
		pthread_cond_signal(&consumer_data.queue_nonempty);
	} else {
		rep->next = consumer_data.head;
		consumer_data.head = rep;
	}

	pthread_mutex_unlock(&consumer_data.queue_lock);
}


static void *event_thread_main(void *arg) 
{
	struct auditd_consumer_data *data = arg;
	unsigned int count = 0L;
	const char *buf = NULL;
	sigset_t sigs;

	/* This is a worker thread. Don't handle signals. */
	sigemptyset(&sigs);
	sigaddset(&sigs, SIGALRM);
	sigaddset(&sigs, SIGTERM);
	sigaddset(&sigs, SIGHUP);
	pthread_sigmask(SIG_BLOCK, &sigs, NULL);

	while (1) {
		struct auditd_reply_list *next;
        
		pthread_mutex_lock(&data->queue_lock);
        
		while (data->head == NULL) {
			pthread_cond_wait(&data->queue_nonempty, 
					&data->queue_lock);
		}

		if (!logging_suspended) {

			switch (data->config->log_format)
			{
				case LF_RAW:
					buf = format_raw(&data->head->reply);
					break;
//				case LF_TEXT:
//					buf = format_text(&data->head->reply);
//					break;
				default:
					msg(LOG_ERR, 
					"Illegal log format detected %d", 
						data->config->log_format);
					break;
			}

			if (buf) {
				write_to_log(buf, data);
				free((void *)buf);
			} else 
				msg(LOG_ERR, "NULL buffer passed to write_log");

			/* See if we need to flush the disk manually */
			if (data->config->flush == FT_INCREMENTAL) {
				count++;
				if ((count % data->config->freq) == 0)
					fflush(data->log_file);
					fsync(data->log_fd);
			}
		}
		next = data->head->next;
		if (data->tail == data->head)
			data->tail = NULL;
		free(data->head);
		data->head = next;
		if (next == NULL && stop) {
			pthread_mutex_unlock(&data->queue_lock);
			return NULL;
		}
		pthread_mutex_unlock(&data->queue_lock);
	}
	return NULL;
}

static int write_to_log(const char *buf, struct auditd_consumer_data *data)
{
	int rc;
	struct stat st;
	FILE *f = data->log_file;
	struct daemon_conf *config = data->config;

	// write it to disk
	rc = fprintf(f, "%s\n", buf);

	// error? Handle it
	if (rc < 0) {
		msg(LOG_ERR, "Record was not written to disk (%s)", 
			strerror(errno));
		// FIXME: Need to notify the kernel
		if (errno == ENOSPC && fs_space_left == 1) {
			fs_space_left = 0;
			do_disk_full_action(config);
		}
		return 1;
	}

	// get file size
	rc = fstat(data->log_fd, &st);
	if (rc == 0) {
		// did we cross the size limit?
		unsigned long sz = st.st_size / MEGABYTE;
		if (sz >= config->max_log_size && 
				(config->daemonize == D_BACKGROUND)) {
			switch (config->max_log_size_action)
			{
				case SZ_IGNORE:
					break;
				case SZ_SYSLOG:
					msg(LOG_ERR,
					    "Audit daemon log file is larger than max size");
					break;
				case SZ_SUSPEND:
					msg(LOG_ERR,
					    "Audit daemon is suspending logging due to logfile size.");
					logging_suspended = 1;
					break;
				case SZ_ROTATE:
					if (data->config->num_logs > 1) {
						msg(LOG_NOTICE,
					    "Audit daemon rotating log files");
						rotate_logs(data);
					}
					break;
				default:
					msg(LOG_ALERT, 
					    "Audit daemon log file is larger than max size and unknown action requested");
					break;
			}
		}
	} else 
		msg(LOG_ERR, "Unable to fstat log file (%s)", strerror(errno));

	// get space left on partition
	if (config->daemonize == D_BACKGROUND)
		check_space_left(data->log_fd, config);
	return 0;
}

static void check_space_left(int lfd, struct daemon_conf *config)
{
	int rc;
	struct statfs buf;

        rc = fstatfs(lfd, &buf);
        if (rc == 0) {
		if (buf.f_bavail < 5) {
			/* we won't consume the last 5 blocks */
			fs_space_left = 0;
			do_disk_full_action(config);
		} else {
			unsigned long blocks;
			unsigned long block_size = buf.f_bsize;
		        blocks = config->space_left * (MEGABYTE/block_size);
        		if (blocks > buf.f_bavail) {
				if (fs_space_warning == 0) {
					fs_space_warning = 1;
					do_space_left_action(config->space_left_action);
				}
			}
		        blocks = config->admin_space_left * (MEGABYTE/block_size);
        		if (blocks > buf.f_bavail) {
				if (fs_admin_space_warning == 0) {
					fs_admin_space_warning = 1;
					do_space_left_action(config->admin_space_left_action);
				}
			}
	
		}
	}
	else msg(LOG_DEBUG, "fstatfs returned:%d, %s", rc, strerror(errno));
}

void do_space_left_action(failure_action_t action)
{
	switch (action)
	{
		case FA_IGNORE:
			break;
		case FA_SYSLOG:
			msg(LOG_ALERT, 
			    "Audit daemon is low on disk space for logging");
			break;
		case FA_SUSPEND:
			msg(LOG_ALERT,
			    "Audit daemon is suspending logging due to low disk space.");
			logging_suspended = 1;
			break;
		case FA_SINGLE:
			msg(LOG_ALERT, 
				"The audit daemon is now changing the system to single user mode");
			(void)system("/sbin/init 1");
			break;
		case FA_HALT:
			msg(LOG_ALERT, 
				"The audit daemon is now halting the system");
			(void)system("/sbin/init 0");
			break;
		default:
			msg(LOG_ALERT, 
			    "Audit daemon is low on disk space for logging and unknown action requested");
			break;
	}
}

void do_disk_full_action(struct daemon_conf *config)
{
	msg(LOG_ALERT, "Audit daemon has no disk space left for logging");
	switch (config->disk_full_action)
	{
		case FA_IGNORE:
		case FA_SYSLOG:
			break;
		case FA_SUSPEND:
			msg(LOG_ALERT,
			    "Audit daemon is suspending logging due to no disk space left.");
			logging_suspended = 1;
			break;
		case FA_HALT:
			msg(LOG_ALERT, 
				"The audit daemon is now halting the system");
			(void)system("/sbin/init 0");
			break;
		default:
			msg(LOG_ALERT, "Unknown disk full action requested");
			break;
	} 
}

static void rotate_logs(struct auditd_consumer_data *data)
{
	int rc;
	unsigned int len, i;
	char *oldname, *newname;

	// Close audit file
	fclose(data->log_file);
	
	// Rotate
	len = strlen(data->config->log_file) + 4;
	oldname = (char *)malloc(len);
	if (oldname == NULL) { /* Not fatal - just messy */
		msg(LOG_ERR, "No memory rotating logs");
		return;
	}
	newname = (char *)malloc(len);
	if (newname == NULL) { /* Not fatal - just messy */
		msg(LOG_ERR, "No memory rotating logs");
		free(oldname);
		return;
	}
	for (i=data->config->num_logs - 1; i>1; i--) {
		int rc;

		snprintf(oldname, len, "%s.%d", data->config->log_file, i-1);
		snprintf(newname, len, "%s.%d", data->config->log_file, i);
		// if the old file exists
		rc = rename(oldname, newname);
		if (rc == -1 && errno != ENOENT) {
			msg(LOG_ERR, "Error rotating logs from %s to %s (%s)",
				oldname, newname, strerror(errno));
		}
	}
	free(newname);

	// At this point, oldname should point to lowest number - use it
	newname = oldname;
	rc = rename(data->config->log_file, newname);
	if (rc == -1 && errno != ENOENT) {
		msg(LOG_ERR, "Error rotating logs from %s to %s (%s)",
			data->config->log_file, newname, strerror(errno));
	}
	free(newname);

	// open new audit file
	if (open_audit_log(data)) {
		msg(LOG_NOTICE, 
			"Could not reopen a log after rotating...exiting.");
		exit(1);
	}
}

/*
 * This function handles opening a descriptor for the audit log
 * file and ensuring the correct options are applied to the descriptor.
 * It returns 0 on success and 1 on failure.
 */
static int open_audit_log(struct auditd_consumer_data *data)
{
	int flags, lfd;

	lfd = open(data->config->log_file, O_WRONLY|O_APPEND|O_NOFOLLOW);
	if (lfd < 0) {
		if (errno == ENOENT) {
			lfd = create_log_file(data->config->log_file);
			if (lfd < 0)
				return 1;
			close(lfd);
			lfd = open(data->config->log_file, 
				O_WRONLY|O_APPEND|O_NOFOLLOW);
		}
		if (lfd < 0) {
			msg(LOG_ERR, "Couldn't open log file %s (%s)",
				data->config->log_file, strerror(errno));
			return 1;
		}
	}

	if (fcntl(lfd, F_SETFD, FD_CLOEXEC) == -1) {
		close(lfd);
		msg(LOG_ERR, "Error setting log file CLOEXEC flag (%s)",
			strerror(errno));
		return 1;
	}
	if (data->config->flush == FT_DATA) {
		flags = fcntl(lfd, F_GETFL);
		if (flags < 0) {
			msg(LOG_ERR, "Couldn't get log file flags (%s)",
				strerror(errno));
			close(lfd);
			return 1;
		}
		if (fcntl(lfd, F_SETFL, flags|O_DSYNC) < 0) {
			msg(LOG_ERR,
				"Couldn't set data sync mode on log file (%s)",
				strerror(errno));
			close(lfd);
			return 1;
		}
	}
	else if (data->config->flush == FT_SYNC){
		flags = fcntl(lfd, F_GETFL);
		if (flags < 0) {
			msg(LOG_ERR, "Couldn't get log file flags (%s)",
				strerror(errno));
			close(lfd);
			return 1;
		}
		if (fcntl(lfd, F_SETFL, flags|O_SYNC) < 0) {
			msg(LOG_ERR,
				"Couldn't set sync mode on log file (%s)",
				strerror(errno));
			close(lfd);
			return 1;
		}
	}

	data->log_fd = lfd;
	data->log_file = fdopen(lfd, "a");
	if (data->log_file == NULL) {
		msg(LOG_ERR, "Error setting up log descriptor (%s)",
			strerror(errno));
		close(lfd);
		return 1;
	}

	/* Set it to line buffering */
	setlinebuf(consumer_data.log_file);
	return 0;
}

